package tagInducer;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import utils.TagStats;

/**
 * The entry-point class.
 * @author Christos Christodoulopoulos
 */
public class Inducer{
	private static String outDir = "results";
	private String outFile, runDir;
	private Corpus corpus;
	private Options o;
	private GibbsSampler sampler;
	private List<String> featureTypes;
	
	public Inducer(String configFile) throws IOException{
		o = new Options(configFile);
		featureTypes = o.getFeatureTypes();
		List<Double> hyperClassList = o.getHyperClassList();
		List<Integer> iterList = o.getIterList();
		List<Map<String, Double>> hyperFeatsList = o.getHyperFeatsList();
		
		//Parse the corpus
		corpus = new Corpus(new File(o.getCorpusFileName()), o.getNumContextFeats());
		corpus.setMorphFile(o.getMorphFile());
		corpus.setMorphMethod(o.getMorphMethod());
		
		if (featureTypes.contains("ALL_CONTEXT")) {
			corpus.addAllContextFeats();
		}
		if (featureTypes.contains("LEFT")) corpus.addLeftContextFeats();
		if (featureTypes.contains("RIGHT")) corpus.addRightContextFeats();
		if (featureTypes.contains("MORPH")) corpus.extendMorph(o.isExtendedMorph());
		if (featureTypes.contains("ALIGNS")) corpus.extendAligns(o.getAlignmentsFile(), 
				o.getCorpusFileName(), o.getAlignLangs());
		
		//DEBUG: Pring the features
//		FileUtils f = new FileUtils();
//		BufferedWriter outFeats = f.createOut(o.getCorpusFileName().
//				substring(0,o.getCorpusFileName().lastIndexOf('.'))+".feats");
//		for (int wordType=0; wordType<corpus.getNumTypes(); wordType++){
//			for (String featType:corpus.getFeatures().keySet()){
//				for (double feat:corpus.getFeatures().get(featType)[wordType]){
//					outFeats.write(Math.round(feat)+" ");
//				}
//			}
//			outFeats.write("\n");
//		}
//		outFeats.close();
//		System.exit(0);
		
		//Construct and configure the sampler
		sampler = new FullGibbsSampler(corpus, o);
		int[] initC = null;
		if (o.isInitWithFreq()) initC = corpus.initFreq(o.getNumClasses());
		sampler.initialise(o.getNumClasses(), initC);
		
		for (int run = 1; run <= o.getNumRuns(); run++){
			int totRuns = hyperFeatsList.size() * hyperClassList.size() * iterList.size();
			int settingsRun = 1;
			for (int iters:iterList){
				for (Map<String, Double> featL:hyperFeatsList){
					for (double classL:hyperClassList){
						System.out.println("[REP: "+run+"/"+o.getNumRuns()+"] - Set "+
								settingsRun+" out of "+totRuns+".");
						runInducer(iters, new HashMap<String, Double>(featL), classL);
						settingsRun++;
						System.out.println();
					}
				}
			}
		}
	}
	
	private void runInducer(int iters, Map<String, Double> hyperFeats, 
			double hyperClass) throws IOException{
		String corpusFileName = o.getCorpusFileName();
		int numClusters = o.getNumClasses();
		boolean initWithFreq = o.isInitWithFreq();
		boolean anneal = o.isAnneal();
		
		//Generate the name of the directory for this run (dName)
		String cName = corpusFileName.substring(corpusFileName.lastIndexOf('/')+1,
				corpusFileName.lastIndexOf('.'));
		String dName = "";
		dName += "c"+numClusters;
		//Add the run's config
		int i = 0;
		for (String type:hyperFeats.keySet()) {
			if (i++ < 1) dName += "_h"+type;
			else dName += "h"+type;
			if (hyperFeats.get(type) != 0.0) dName += hyperFeats.get(type);
			else dName += "+S";
		}
		if (hyperClass != 0.0) dName +="_hClass"+hyperClass;
		else dName +="_hClass+S";
		if (featureTypes.contains("ALIGNS")){
			if (o.getAlignLangs() != null) {
				dName += "_aligns";
				for (String lang:o.getAlignLangs()) dName +=lang.toUpperCase();
			}
			else {
				System.err.println("Aligns feature types chosen without alignment languages!");
				System.exit(1);
			}
		}
		if (featureTypes.contains("MORPH")) {
			if (o.getMorphMethod().equalsIgnoreCase("morfessor")) dName += "_morphM"; 
			else if (o.getMorphMethod().equalsIgnoreCase("letter")) dName += "_morphL";
		}
		if(initWithFreq) dName+="_initC";
		if(featureTypes.contains("MORPH") && o.isExtendedMorph()) dName+="_extMFeats";
		if (anneal) dName += "_anneal";
		if (o.getTiedHyperType()!=null) dName += "_"+o.getTiedHyperType();
		dName += "_i"+iters;
		
		int maxRun = 0;
		File dFile = new File(outDir+"/"+dName);
		if (dFile.isDirectory()){
			File[] subDirs = dFile.listFiles();
			for (File subDir:subDirs){
				String sDirName = subDir.getName();
				if (sDirName.startsWith(cName)){
					int run = Integer.parseInt(sDirName.substring(
							sDirName.lastIndexOf('_')+1,sDirName.length()));
					if (run > maxRun) maxRun = run;
				}
			}
		}
		maxRun++;
		dName += "/"+cName+"_"+maxRun;
		
		//Create the output dir for this run
		runDir = outDir+"/"+dName;
		if (!new File(runDir).mkdirs()) {
			System.err.println("Couldn't create output dirs"); 
			System.exit(1);
		}
		

		//Set the parameters for the current run
		sampler.setRunParameters(iters, hyperFeats, hyperClass);
		
		//**Run the sampler!**
		sampler.gibbs(runDir);

		//Output the tagged file
		outFile = runDir+"/out.tagged";
		corpus.writeTagged(sampler.getTokenAssignments(true), outFile);
		TagStats stat = new TagStats();
		//Do some statistics
		if (corpus.hasTags()){
			stat.createTagDist(outFile);
			stat.createTagMap(corpusFileName, outFile);
			stat.createConfusionMatrix(corpusFileName, outFile);
			stat.createClusters(outFile, 20);
		}
	}

	public static void main(String[] args) {
		try{new Inducer(args[0]);}
		catch(IOException e){e.printStackTrace();}
	}
}
