package tagInducer;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import utils.MapUtil;

/**
 * A storage class to keep all the sampler options.
 * @author Christos Christodoulopoulos
 */
public class Options {
	private static Properties config = new Properties();
	private int numClasses, numRuns, numContextFeats;
	private boolean printLog, anneal, initWithFreq, extendedMorph;
	private String morphMethod, untiedHyperType, morphFile, alignmentsFile, corpusFileName;
	private String[] alignLangs;
	private List<Integer> iterList = new ArrayList<Integer>();
	private List<String> featureTypes = new ArrayList<String>();
	private List<Double> hyperClassList = new ArrayList<Double>();
	private List<Map<String, Double>> hyperFeatsList = new ArrayList<Map<String, Double>>();

	public Options(String configFile){
		String temp;
		try {
			FileInputStream in = new FileInputStream(configFile);
			config.load(in); 
		}
		catch (IOException  e) {
			System.err.println("Cannot read configuration file " + configFile);
			System.exit(1);
		}
		corpusFileName = config.getProperty("CORPUS");
		numClasses = Integer.parseInt(config.getProperty("NUM_CLUSTERS"));
		temp = config.getProperty("FEATURE_TYPES");
		if (temp != null){
			if (temp.contains("[")){
				String[] types = temp.substring(1, temp.length()-1).split("\\s+");
				for (String type:types) featureTypes.add(type.toUpperCase());
			}
			else featureTypes.add(temp.toUpperCase());
		}
		initWithFreq = Boolean.parseBoolean(config.getProperty("INTIALISE_W_FREQUENCY"));
		extendedMorph = Boolean.parseBoolean(config.getProperty("EXTENDED_MORPH"));
		
		//Get feature hyperparameters - first into a Map of Lists
		Map<String, List<Double>> hyperTemp = new HashMap<String, List<Double>>();
		getFeatPrior("HYPER_ALL_CONTEXT", hyperTemp);
		getFeatPrior("HYPER_LEFT", hyperTemp);
		getFeatPrior("HYPER_RIGHT", hyperTemp);
		getFeatPrior("HYPER_MORPH", hyperTemp);
		getFeatPrior("HYPER_ALIGNS", hyperTemp);
		//and then turn it into a List of Maps
		hyperFeatsList = MapUtil.MapList2ListMap(hyperTemp);
		//END: Get feature hyperparameters
		
		temp = config.getProperty("HYPER_CLASS");
		if (temp != null){
			if (temp.contains("[")){
				hyperClassList = new ArrayList<Double>();
				String[] nums = temp.substring(1, temp.length()-1).split("\\s+");
				for (String num:nums) hyperClassList.add(Double.parseDouble(num));
			}
			else hyperClassList.add(Double.parseDouble(temp));
		}
		else hyperClassList.add(0.0);
		
		temp = config.getProperty("SAMPLE_ITERS");
		if (temp.contains("[")){
			String[] nums = temp.substring(1, temp.length()-1).split("\\s+");
			for (String num:nums) iterList.add(Integer.parseInt(num));
		}
		else iterList.add(Integer.parseInt(temp));
		untiedHyperType = config.getProperty("UNTIED_HYPER_TYPE");
		temp = config.getProperty("NUM_RUNS");
		if (temp != null) numRuns = Integer.parseInt(temp);
		temp = config.getProperty("NUM_CONTEXT_FEATS");
		if (temp != null) numContextFeats = Integer.parseInt(temp);
		printLog = Boolean.parseBoolean(config.getProperty("PRINT_LOG"));
		anneal = Boolean.parseBoolean(config.getProperty("ANNEALING"));
		morphMethod = config.getProperty("MORPH_METHOD");
		morphFile = config.getProperty("MORPH_FILE");
		alignmentsFile = config.getProperty("ALIGNMENTS_FILE");
		temp = config.getProperty("ALIGNMENT_LANGS");
		if (temp != null){
			if (temp.contains("[")){
				alignLangs = temp.substring(1, temp.length()-1).split("\\s+");
			}
			else {
				alignLangs = new String[1];
				alignLangs[0] = temp;
			}
		}
	}

	private void getFeatPrior(String priorString, Map<String, List<Double>> hyperTemp) {
		String temp = config.getProperty(priorString);
		String featString = priorString.substring(priorString.indexOf('_')+1);
		List<Double> tempList = new ArrayList<Double>();
		if (temp != null){
			if (temp.contains("[")){
				String[] nums = temp.substring(1, temp.length()-1).split("\\s+");
				for (String num:nums) tempList.add(Double.parseDouble(num));
			}
			else tempList.add(Double.parseDouble(temp));
			hyperTemp.put(featString, tempList);
		}
		else if (featureTypes.contains(featString)) {
			tempList.add(0.0);
			hyperTemp.put(featString, tempList);
		}
	}
	
	public int getNumClasses() {return numClasses;}
	public int getNumRuns() {return numRuns;}
	public int getNumContextFeats() {return numContextFeats;}
	public boolean isPrintLog() {return printLog;}
	public boolean isAnneal() {return anneal;}
	public boolean isInitWithFreq() {return initWithFreq;}
	public boolean isExtendedMorph() {return extendedMorph;}
	public String getMorphMethod() {return morphMethod;}
	public String getMorphFile() {return morphFile;}
	public String getAlignmentsFile() {return alignmentsFile;}
	public String getCorpusFileName() {return corpusFileName;}
	public String[] getAlignLangs() {return alignLangs;}
	public List<Integer> getIterList() {return iterList;}
	public List<String> getFeatureTypes() {return featureTypes;}
	public List<Double> getHyperClassList() {return hyperClassList;}
	public List<Map<String, Double>> getHyperFeatsList() {return hyperFeatsList;}
	public String getTiedHyperType() {return untiedHyperType;}
}
