package utils;

public class Gamma {

	/** Lanczos coefficients */
	private static double[] lanczos =
	{
		0.99999999999999709182,
		57.156235665862923517,
		-59.597960355475491248,
		14.136097974741747174,
		-0.49191381609762019978,
		.33994649984811888699e-4,
		.46523628927048575665e-4,
		-.98374475304879564677e-4,
		.15808870322491248884e-3,
		-.21026444172410488319e-3,
		.21743961811521264320e-3,
		-.16431810653676389022e-3,
		.84418223983852743293e-4,
		-.26190838401581408670e-4,
		.36899182659531622704e-5,
	};

	/** Avoid repeated computation of log of 2 PI in logGamma */
	private static final double HALF_LOG_2_PI = 0.5 * Math.log(2.0 * Math.PI);

	/**
	 * Returns the natural logarithm of the gamma function &#915;(x).
	 *
	 * The implementation of this method is based on:
	 * <ul>
	 * <li><a HREF="http://mathworld.wolfram.com/GammaFunction.html">
	 * Gamma Function</a>, equation (28).</li>
	 * <li><a HREF="http://mathworld.wolfram.com/LanczosApproximation.html">
	 * Lanczos Approximation</a>, equations (1) through (5).</li>
	 * <li><a HREF="http://my.fit.edu/~gabdo/gamma.txt">Paul Godfrey, A note on
	 * the computation of the convergent Lanczos complex Gamma approximation
	 * </a></li>
	 * </ul>
	 *
	 * @param x the value.
	 * @return log(&#915;(x))
	 */
	public static double logGamma(double x) {
		double ret;

		if (Double.isNaN(x) || (x <= 0.0)) {
			ret = Double.NaN;
		} else {
			double g = 607.0 / 128.0;

			double sum = 0.0;
			for (int i = lanczos.length - 1; i > 0; --i) {
				sum = sum + (lanczos[i] / (x + i));
			}
			sum = sum + lanczos[0];

			double tmp = x + g + .5;
			ret = ((x + .5) * Math.log(tmp)) - tmp +
			HALF_LOG_2_PI + Math.log(sum / x);
		}

		return ret;
	}
	
	public static double logGamma2(double x) {
	      double tmp = (x - 0.5) * Math.log(x + 4.5) - (x + 4.5);
	      double ser = 1.0 + 76.18009173    / (x + 0)   - 86.50532033    / (x + 1)
	                       + 24.01409822    / (x + 2)   -  1.231739516   / (x + 3)
	                       +  0.00120858003 / (x + 4)   -  0.00000536382 / (x + 5);
	      return tmp + Math.log(ser * Math.sqrt(2 * Math.PI));
	   }
}